import string, xml.sax

__version__ = "$Revision: 1.4 $"

class ServerReader(xml.sax.ContentHandler):

    def setDocumentLocator(self, locator):
        self.__locator = locator

    def startDocument(self):
        """Initialize self by creating empty private variables used in parsing"""

        # this member holds information about the server element being parsed
        self.__server = {}

        # this member holds a dictionary with all the servers that have been
        # completely parsed so far
        self.__serverList = {}

        # this holds the char data currently being read
        self.__charData = ''

        # __beginElements and __endElements are dictionaries to facilitate parsing
        # the configuration file.  Basically they serve as dictionaries that
        # map elements to the functions that handle them
        self.__beginElements = {'serverlist' : self.__startServerListElement,
                                'server' : self.__startServerElement,
                                'address' : self.__startAddressElement,
                                'path' : self.__startPathElement}

        self.__endElements = {'serverlist' : self.__endServerListElement,
                              'server' : self.__endServerElement,
                              'address' : self.__endAddressElement,
                              'path' : self.__endPathElement}

        # indicates that file is being parsed
        self.__parsing = 1

    def endDocument(self):
        """Indicate that parsing of the document is done"""

        self.__parsing = 0
        
    def startElement(self, name, attributes):
        """Handle start element events

        Works by checking to see if the element is in the __beginElements
        dictionary.   If so then it calls the function associated with that
        key with the element's attributes.  If the element is not in the
        dictionary an InvalidXML exception is raised"""

        if (self.__beginElements.has_key(name)):
            self.__beginElements[name](attributes)
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def endElement(self, name):
        """Handle end element events

        Works by checking to see if the element is in the __endElements
        dictionary.   If so then it calls the function associated with that
        key with the element's attributes.  If the element is not in the
        dictionary an InvalidXML exception is raised"""

        if (self.__endElements.has_key(name)):
            self.__endElements[name]()
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def __startAddressElement(self, attributes):
        """Handle the beginning of an address element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: Address elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endAddressElement(self):
        """Handle the end of an address element by setting the address key of
        self.__server.  Raises an InvaldXML exception if attributes
        is not empty or if nothing is contained within the address element"""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: Address elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__server['address'] = string.strip(self.__charData)
        
    def __startPathElement(self, attributes):
        """Handle the beginning of an path element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: Path elements don't have attributes!" %
                             (self.__locator.getLineNumber()))
    
    def __endPathElement(self):
        """Handle the end of an path element by setting the path key of
        self.__server.  Raises an InvaldXML exception if attributes
        is not empty"""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: Path elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__server['path'] = string.strip(self.__charData)

    def __startServerElement(self, attributes):
        """Handle the beginning of an server element by clearing the
        self.__server member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__server = {}
        else:
            raise InvalidXML("l. %d: Server elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endServerElement(self):
        """Handle the end of an server element by adding the
        self.__server member to self.__serverlist.  Raises an InvaldXML exception
        if attributes is not empty or self.__server doesn't have exactly 2 keys"""

        keys = self.__server.keys()
        
        if 'address' in keys and 'path' in keys and len(keys) == 2:
            self.__serverList.append(self.__server)

    def __startServerListElement(self, attributes):
        """Handle the beginning of an servers element by clearing the
        self.__serverList member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__serverList = []
        else:
            raise InvalidXML("l. %d: Servers elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endServerListElement(self):
        """Handle the end of an servers element.  Raises an InvalidXML exception
        if attributes is non-empty, otherwise does nothing"""
        pass
    
    def characters(self, content):
        """Append characters to self.__charData"""

        self.__charData += content

    def getServerList(self):
        """Returns serverlist if parsing is complete, raises a NotAllowed
        exception otherwise"""

        if self.__parsing:
            raise NotAllowed("Can't get serverlist while parsing xml")
        else:
            return self.__serverList
        

class PackageReader(xml.sax.ContentHandler):

    def setDocumentLocator(self, locator):
        self.__locator = locator

    def startDocument(self):
        """Initialize self by creating empty private variables used in parsing"""

        # this member holds information about the package element being parsed
        self.__package = {}
        self.__dependencies = {}
        self.__inDependency = 0
        self.__dependencyPackage = {}
        self.__authors = []

        # this member holds a dictionary with all the pacakges that have been
        # completely parsed so far
        self.__packageList = {}

        # this holds the char data currently being read
        self.__charData = ''

        # __beginElements and __endElements are dictionaries to facilitate parsing
        # the configuration file.  Basically they serve as dictionaries that
        # map elements to the functions that handle them
        self.__beginElements = {'packagelist' : self.__startPackageListElement,
                                'package' : self.__startPackageElement,
                                'filename' : self.__startFilenameElement,
                                'name' : self.__startNameElement,
                                'version' : self.__startVersionElement,
                                'author' : self.__startAuthorElement,
                                'license' : self.__startLicenseElement,
                                'description' : self.__startDescriptionElement,
                                'path' : self.__startPathElement,
                                'url' : self.__startUrlElement,
                                'size' : self.__startSizeElement,
                                'dependency' : self.__startDependencyElement,
                                'dependencies' : self.__startDependenciesElement}

        self.__endElements = {'packagelist' : self.__endPackageListElement,
                              'package' : self.__endPackageElement,
                              'name' : self.__endNameElement,
                              'filename' : self.__endFilenameElement,
                              'version' : self.__endVersionElement,
                              'author' : self.__endAuthorElement,
                              'license' : self.__endLicenseElement,
                              'description' : self.__endDescriptionElement,
                              'url' : self.__endUrlElement,
                              'path' : self.__endPathElement,
                              'size' : self.__endSizeElement,
                              'dependency' : self.__endDependencyElement,
                              'dependencies' : self.__endDependenciesElement}

        # indicates that file is being parsed
        self.__parsing = 1

    def endDocument(self):
        """Indicate that parsing of the document is done"""

        self.__parsing = 0
        
    def startElement(self, name, attributes):
        """Handle start element events

        Works by checking to see if the element is in the __beginElements
        dictionary.   If so then it calls the function associated with that
        key with the element's attributes.  If the element is not in the
        dictionary an InvalidXML exception is raised"""

        if (self.__beginElements.has_key(name)):
            self.__beginElements[name](attributes)
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def endElement(self, name):
        """Handle end element events

        Works by checking to see if the element is in the __endElements
        dictionary.   If so then it calls the function associated with that
        key with the element's attributes.  If the element is not in the
        dictionary an InvalidXML exception is raised"""

        if (self.__endElements.has_key(name)):
            self.__endElements[name]()
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def __startAuthorElement(self, attributes):
        """Handle the beginning of an author element by setting the self.__inAuthor
        for __startNameElement and __endNameElement methods. Raises an InvaldXML
        exception if attributes is not empty"""

        if attributes.getLength() == 2:
            name = attributes.getValue('name')
            email = attributes.getValue('email')
            self.__authors.append((name, email))
        else:
            raise InvalidXML("l. %d: Address elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endAuthorElement(self):
        """Handle the end of an author element, dummy function for now"""
        pass
        
    def __startDescriptionElement(self, attributes):
        """Handle the beginning of an description element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: description elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endDescriptionElement(self):
        """Handle the end of an description element by setting the description
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: description elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__package['description'] = string.strip(self.__charData)

    def __startLicenseElement(self, attributes):
        """Handle the beginning of an license element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: license elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endLicenseElement(self):
        """Handle the end of an license element by setting the license
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: license elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__package['license'] = string.strip(self.__charData)

    def __startVersionElement(self, attributes):
        """Handle the beginning of an version element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty unless we are processing dependency information"""

        if attributes.getLength() == 0 and not self.__inDependency:
            self.__charData = ''
        elif self.__inDependency and attributes.getLength() == 1:
            self.__dependencyPackage['version-type'] = attributes.getValue('type')
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: version elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endVersionElement(self):
        """Handle the end of an version element. If we are processing 
        dependency information update self.__dependencyPackage set
        update self.__package['version']"""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: version elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        if self.__inDependency:
            self.__dependencyPackage['version'] = string.strip(self.__charData)
        else:
            self.__package['version'] = string.strip(self.__charData)

    def __startUrlElement(self, attributes):
        """Handle the beginning of an url element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: url elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endUrlElement(self):
        """Handle the end of an url element by setting the url
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: url elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__package['url'] = string.strip(self.__charData)

    def __startFilenameElement(self, attributes):
        """Handle the beginning of an filename element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: filename elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endFilenameElement(self):
        """Handle the end of an filename element by setting the filename
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: filename elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__package['filename'] = string.strip(self.__charData)

    def __startSizeElement(self, attributes):
        """Handle the beginning of a size element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: size elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endSizeElement(self):
        """Handle the end of a size element by setting the size
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: size elements must contain a number!" %
                             (self.__locator.getLineNumber()))

        try:
            size = int(string.strip(self.__charData))
            self.__package['size'] = size
        except TypeError:
            raise InvalidXML("l. %d: size elements must contain a number!" %
                             (self.__locator.getLineNumber()))
            

    def __startPathElement(self, attributes):
        """Handle the beginning of an path element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: path elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endPathElement(self):
        """Handle the end of an path element by setting path key of
        self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: path elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        
        self.__package['path'] = string.strip(self.__charData)

    def __startNameElement(self, attributes):
        """Handle the beginning of an name element by clearing the
        self.__charData member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__charData = ''
        else:
            raise InvalidXML("l. %d: name elements don't have attributes"
                             % (self.__locator.getLineNumber()))
    
    def __endNameElement(self):
        """Handle the end of an name element.  If processing dependency information
        set self.__dependencyPackage['name'] otherwiise set the name
        key of self.__package."""

        if len(self.__charData) == 0:
            raise InvalidXML("l. %d: name elements must contain a string!" %
                             (self.__locator.getLineNumber()))
        if self.__inDependency:
            self.__dependencyPackage['name'] = string.strip(self.__charData)
        else:
            self.__package['name'] = string.strip(self.__charData)


    def __startDependencyElement(self, attributes):
        """Handle the beginning of an dependency element. Raises an InvaldXML
        exception if attributes is not empty"""

        if attributes.getLength() == 1:
            self.__dependencyPackage = {}
            self.__inDependency = 1
            self.__dependencyPackage['type'] = attributes.getValue('type')
        else:
            raise InvalidXML("l. %d: dependency elements require a type attrbute!"
                             % (self.__locator.getLineNumber()))
    
    def __endDependencyElement(self):
        """Handle the end of an dependency element by adding the completed
        dependency package information to self.__dependencies"""
        dependencyName = self.__dependencyPackage['name']
        self.__dependencies[dependencyName] = self.__dependencyPackage

    def __startDependenciesElement(self, attributes):
        """Handle the beginning of a dependencies element by clearing the
        self.__dependencies member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__dependencies = {}
        else:
            raise InvalidXML("l. %d: dependencies elements don't have attributes" %
                             (self.__locator.getLineNumber()))

    def __endDependenciesElement(self):
        """Handle the end of an dependencies element by adding the
        self.__dependencies member to self.__package."""

        self.__package['dependencies'] = self.__dependencies


    def __startPackageElement(self, attributes):
        """Handle the beginning of an package element by clearing the
        self.__package member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__package = {}
            self.__authors = []
        else:
            raise InvalidXML("l. %d: package elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endPackageElement(self):
        """Handle the end of an package element. Add authors to self.__package
        and then add package to self.__packagelist"""

        self.__package['author'] = self.__authors
        self.__packageList[self.__package['filename']] = self.__package

    def __startPackageListElement(self, attributes):
        """Handle the beginning of a packagelist element by clearing the
        self.__packageList member.  Raises an InvaldXML exception if attributes
        is not empty"""

        if attributes.getLength() == 0:
            self.__packageList = {}
        else:
            raise InvalidXML("l. %d: packagelist elements don't have attributes!" %
                             (self.__locator.getLineNumber()))

    def __endPackageListElement(self):
        """Handle the end of an packagelist element.  Raises an InvalidXML
        exception if attributes is non-empty, otherwise does nothing"""
        pass
    
    def characters(self, content):
        """Append characters to self.__charData"""

        self.__charData += content

    def getPackageList(self):
        """Returns serverlist if parsing is complete, raises a NotAllowed
        exception otherwise"""

        if self.__parsing:
            raise NotAllowed("Can't get packagelist while parsing xml")
        else:
            return self.__packageList
        

class InstalledReader(xml.sax.ContentHandler):

    def setDocumentLocator(self, locator):
        self.__locator = locator

    def startDocument(self):
        """Initialize self by creating empty private variables used in parsing"""


        # this member holds a dictionary with all the packages that have been
        # completely parsed so far
        self.__installedList = {}

        # indicates that file is being parsed
        self.__parsing = 1

    def endDocument(self):
        """Indicate that parsing of the document is done"""

        self.__parsing = 0
        
    def startElement(self, name, attributes):
        """Handle start element events"""

        if name == "package":
            if attributes.has_key('name') and attributes.has_key('version'):
                self.__installedList[attributes.getValue('name')] = attributes.getValue('version')
            else:
                raise InvalidXML("l. %d: Missing name or version attribute" %
                                 (self.__locator.getLineNumber()))
        elif name == "packagelist":
            pass
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def endElement(self, name):
        """Handle end element events"""

        if name == 'package' or name == 'packagelist':
            pass
        else:
            raise InvalidXML("l. %d: Invalid element!" %
                             (self.__locator.getLineNumber()))

    def getInstalledList(self):
        """Returns installedList if parsing is complete, raises a NotAllowed
        exception otherwise"""

        if self.__parsing:
            raise NotAllowed("Can't get serverlist while parsing xml")
        else:
            return self.__installedList


# ******** Exceptions thrown by serverReader and packageReader classes ******

class InvalidXML(Exception):
    """Indicates invalid or malformed xml configuration file"""
    pass

class NotAllowed(Exception):
    """Indicates invalid or malformed xml configuration file"""
    pass



